@@ -92,7 +92,9 @@ static gboolean gok_spy_is_visible (Accessible *accessible);
 		
 static void gok_spy_free_nodes (GSList *nodelist);										
 static void gok_spy_resolve_namesakes (GSList *nodelist);
-												
+
+static gboolean gok_spy_is_primary_container (Accessible *accessible);
+									
 /*
  * gok-spy variables
  */
@@ -327,6 +329,10 @@ void gok_spy_open()
 			objectStateListener, "object:state-changed");
 		gok_log ("SPI_registerGlobalEventListener for object:state-changed returned: %d", success);
 
+		success = SPI_registerGlobalEventListener (
+			objectStateListener, "object:children-changed");
+		gok_log ("SPI_registerGlobalEventListener for object:children-changed returned: %d", success);
+
 	}
 	gok_log_leave ();
 }
@@ -560,6 +566,22 @@ gok_spy_node_match (AccessibleNode *node, GokSpySearchType type)
 	case GOK_SPY_SEARCH_ALL:
 		return TRUE;
 		break;
+	case GOK_SPY_SEARCH_ACTIONABLE:
+		if (Accessible_isAction (node->paccessible))
+		{
+			return TRUE;
+		}
+		else 
+		{
+		    AccessibleStateSet *state = Accessible_getStateSet (node->paccessible);
+		    if (AccessibleStateSet_contains (state, SPI_STATE_SELECTABLE))
+		    {
+			AccessibleStateSet_unref (state);
+			return TRUE;
+		    }
+		    AccessibleStateSet_unref (state);
+		}
+		break;
 	case GOK_SPY_SEARCH_APPLICATIONS:
 		if (Accessible_isApplication (node->paccessible))
 			return TRUE;
@@ -690,6 +712,47 @@ gok_spy_get_children (Accessible *search_root)
 	return nodes;
 }
 
+/**
+ * gok_spy_get_actionable_descendants:
+ * Returns a list of actionable or selectable descendants of #accessible.
+ **/
+GSList *
+gok_spy_get_actionable_descendants (Accessible *accessible, GSList *nodes)
+{
+    gint child_count, i, max_children = 20;
+    Accessible *child;
+    AccessibleNodeFlags flags;
+    GSList *ret_nodes = nodes;
+    gboolean is_selection = Accessible_isSelection (accessible);
+    flags.value = 0;
+    flags.data.is_ui = TRUE;
+    child_count = Accessible_getChildCount (accessible);
+    for (i = 0; i < child_count && i < max_children; ++i) 
+    {
+	AccessibleStateSet *stateset;
+	child = Accessible_getChildAtIndex (accessible, i);
+	stateset = Accessible_getStateSet (child);
+	if (Accessible_isAction (child) && AccessibleStateSet_contains (stateset, SPI_STATE_SHOWING))
+	{
+	    ret_nodes = gok_spy_append_node (ret_nodes, child, flags);
+	}
+	else if (is_selection)
+	{
+	    if (AccessibleStateSet_contains (stateset, SPI_STATE_SELECTABLE))
+	    {
+		ret_nodes = gok_spy_append_node (ret_nodes, child, flags);
+	    }
+	}
+	else
+	{
+	    ret_nodes = gok_spy_get_actionable_descendants (child, ret_nodes);
+	}
+	AccessibleStateSet_unref (stateset);
+	Accessible_unref (child);
+    }
+    return ret_nodes;
+}
+
 /** 
 * gok_spy_get_list 
 * 
@@ -770,6 +833,7 @@ gboolean gok_spy_check_queues(void)
 {
 	EventNode* en = NULL;
 	char *s;
+	GokKeyboard *current_kbd;
 
 	gok_log_enter();
 	
@@ -803,7 +867,7 @@ gboolean gok_spy_check_queues(void)
 		{
 		case GOKSPY_FOCUS_EVENT:
 			/* TODO: remove obsolete focus events from queue since process
-			   accessible is expensive. */
+			   focus is expensive. */
 			gok_log("GOKSPY_FOCUS_EVENT");
 			gok_spy_process_focus (en->event->source);
 			break;
@@ -832,6 +896,13 @@ gboolean gok_spy_check_queues(void)
 			m_ptheAccessibleWithText = NULL;
 			((AccessibleChangeListenerCB)(m_ptheWindowChangeListener))(m_ptheWindowAccessible);
 			break;
+		case GOKSPY_CONTAINER_EVENT:
+			gok_log("GOKSPY_CONTAINER_EVENT");
+			current_kbd = gok_main_get_current_keyboard ();
+			gok_spy_update_component_list (gok_main_get_foreground_window_accessible (), 
+			  current_kbd->flags);
+			gok_main_ds (current_kbd);
+			break;
 		case GOKSPY_WINDOW_DEACTIVATE_EVENT:
 			gok_log("GOKSPY_WINDOW_DEACTIVATE_EVENT");
 
@@ -998,7 +1069,7 @@ gok_spy_is_ui (Accessible *accessible, AccessibleRole role)
 	default:
 		if (Accessible_isEditableText (accessible) ||
 		    Accessible_isHypertext (accessible) ||
-		    Accessible_isAction (accessible))
+		    (Accessible_isAction (accessible) && !gok_spy_is_menu_role (role)))
 			interesting = TRUE;
 		gok_log ("checking for interesting interfaces... %s\n", interesting ? "yes" : "no");
 		gok_log ("is icon: %s\n", role == SPI_ROLE_ICON ? "yes" : "no");
@@ -1283,19 +1354,20 @@ GSList* gok_spy_append_node (GSList* nodes,
 			Accessible *parent = Accessible_getParent (pAccessible);
 			/* if there are no siblings, we can use an ancestor's name */
 			gok_log("checking ancestor names.");
-			maxloops = 5; /* ignore great great great great grandparent */
+			maxloops = 7; /* ignore great great great great great great grandparent */
 			while (parent && Accessible_getChildCount (parent) == 1) 
 			{
 				char *tmp;
-				if (maxloops-- < 1) { break; }  /* guard against erroneous cyclic parent/child relations */
 				Accessible *tmp_parent = parent;
+				if (maxloops-- < 1) { break; }  /* guard against erroneous cyclic parent/child relations */
 				tmp = Accessible_getName (tmp_parent); 
+				gok_log ("getting parent name: parent %x, name %s, %d more loops allowed.\n", tmp_parent, tmp ? tmp : "", maxloops);
 				if (tmp != NULL) 
 				{
 				    if (strlen (tmp) > 0) 
 				    {
 					pName = g_strdup (tmp);
-				    SPI_freeString (tmp);
+					SPI_freeString (tmp);
 					break;
 				    }
 				}
@@ -1328,6 +1400,39 @@ GSList* gok_spy_append_node (GSList* nodes,
 				SPI_freeString (s);
 			}
 		}
+
+		if (pName == NULL) { /* desperate attempt to reuse singleton-child's name */
+			Accessible *tmp_parent = pAccessible;
+			gok_log("no description... checking for singleton child with name.");
+			maxloops = 3;
+			Accessible_ref (tmp_parent);
+			while (tmp_parent && Accessible_getChildCount (tmp_parent) == 1 && maxloops) 
+			{
+				Accessible *child;
+				child = Accessible_getChildAtIndex (tmp_parent, 0);
+				if (child) 
+				{
+					s = Accessible_getName (child);
+					if (s)
+					{
+						gint len = strlen (s);
+						if (len > 0) {
+							pName = g_strndup (s, len);
+							Accessible_unref (tmp_parent);
+							tmp_parent = NULL;
+						}
+						SPI_freeString (s);
+					}
+				}
+				if (tmp_parent) 
+				{
+					Accessible_unref (tmp_parent);
+					tmp_parent = child;
+				}
+				--maxloops;
+			}
+		}
+
 		if (pName != NULL) {
 			gok_log("we have a name: [%s]",pName);
 			/* create a new node */
@@ -1466,6 +1571,15 @@ gok_spy_keymap_listener (void)
 	return;
 }
 
+
+gboolean
+gok_spy_is_primary_container (Accessible *accessible)
+{
+    AccessibleRole role;
+    if (accessible) role = Accessible_getRole (accessible);
+    return (role == SPI_ROLE_HTML_CONTAINER || role == SPI_ROLE_SCROLL_PANE);
+}
+
 /** 
 * gok_spy_defunctness_listener 
 *
@@ -1484,6 +1598,15 @@ void gok_spy_defunctness_listener (const AccessibleEvent *event, void *user_data
 		gok_spy_add_idle_handler ();
 		gok_log ("Our target window is now defunct!");
 	}
+	else if (gok_spy_is_primary_container (event->source)) 
+	{
+		en = (EventNode*) g_malloc(sizeof(EventNode));
+		en->event = NULL;
+		en->type = GOKSPY_CONTAINER_EVENT;
+		g_queue_push_tail ( eventQueue, en );
+		gok_spy_add_idle_handler ();
+		gok_log ("A primary UI container is now defunct!");
+	}
 	
 	gok_log_leave();
 }
@@ -1522,25 +1645,33 @@ gok_spy_window_deactivate_listener(const AccessibleEvent *event,
 	gok_log_leave ();
 }
 
-
 static void
 gok_spy_object_state_listener(const AccessibleEvent *event,
                                  void *user_data)
 {
 	EventNode *en;
+	gboolean is_container_add = !strcmp (event->type, "object:children-changed:add");
+
 	gok_log_enter ();
 
+	if (is_container_add)
+	{
+	    if (!event->source || !gok_spy_is_primary_container (event->source))
+	    {
+		gok_log_leave ();
+		return;
+	    }
+	}
 	AccessibleEvent_ref (event);
 	en = (EventNode*)g_malloc (sizeof (EventNode));
 	en->event = event;
-	en->type = GOKSPY_STATE_EVENT;
+	en->type = is_container_add ? GOKSPY_CONTAINER_EVENT : GOKSPY_STATE_EVENT;
 	g_queue_push_tail (eventQueue, en);
 	gok_spy_add_idle_handler ();
 
 	gok_log_leave ();
 }
 
-
 void
 gok_spy_set_context_menu_accessible (Accessible *accessible, int action_index)
 {
@@ -1908,9 +2039,12 @@ gok_spy_search_component_list (Accessible *rootAccessible, GokSpyUIFlags keyboar
 					Accessible_unref (toolbar_item);
 				}
 			}
-			else if (keyboard_ui_flags.data.gui && !flags.data.is_menu && gok_spy_is_ui (child, role))
+			else if ((keyboard_ui_flags.data.gui || keyboard_ui_flags.data.editable_text) && !flags.data.is_menu && gok_spy_is_ui (child, role))
 			{
-				ui_flags.data.gui = flags.data.is_ui = TRUE;
+				if (keyboard_ui_flags.data.gui) 
+				{
+				    ui_flags.data.gui = flags.data.is_ui = TRUE;
+				}
 				if (keyboard_ui_flags.data.editable_text && Accessible_isEditableText (child)) 
 				{
 					ui_flags.data.editable_text = TRUE;
